import { createElement, Component, createContext } from 'react';
import { oneOf, func, bool, number, string, object } from 'prop-types';
import cnb from 'cnbuilder';
import { DraggableCore } from 'react-draggable';

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function _objectWithoutProperties(source, excluded) {
  if (source == null) return {};

  var target = _objectWithoutPropertiesLoose(source, excluded);

  var key, i;

  if (Object.getOwnPropertySymbols) {
    var sourceSymbolKeys = Object.getOwnPropertySymbols(source);

    for (i = 0; i < sourceSymbolKeys.length; i++) {
      key = sourceSymbolKeys[i];
      if (excluded.indexOf(key) >= 0) continue;
      if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue;
      target[key] = source[key];
    }
  }

  return target;
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

/**
 * @description Will return renderer result if presented, div element otherwise.
 * If renderer is presented it'll receive `elementRef` function which should be used as HTMLElement's ref.
 *
 * @param props {ElementPropsWithElementRefAndRenderer}
 * @param elementRef {ElementRef}
 */

function renderDivWithRenderer(props, elementRef) {
  if (props.renderer) {
    props.elementRef = elementRef;
    var renderer = props.renderer;
    delete props.renderer;
    return renderer(props);
  }

  return createElement("div", Object.assign({}, props, {
    ref: elementRef
  }));
}
var AXIS_DIRECTION;

(function (AXIS_DIRECTION) {
  AXIS_DIRECTION["X"] = "x";
  AXIS_DIRECTION["Y"] = "y";
})(AXIS_DIRECTION || (AXIS_DIRECTION = {}));

var AXIS_DIRECTION_PROP_TYPE = oneOf([AXIS_DIRECTION.X, AXIS_DIRECTION.Y]);
var TRACK_CLICK_BEHAVIOR;

(function (TRACK_CLICK_BEHAVIOR) {
  TRACK_CLICK_BEHAVIOR["JUMP"] = "jump";
  TRACK_CLICK_BEHAVIOR["STEP"] = "step";
})(TRACK_CLICK_BEHAVIOR || (TRACK_CLICK_BEHAVIOR = {}));

var TRACK_CLICK_BEHAVIOR_PROP_TYPE = oneOf([TRACK_CLICK_BEHAVIOR.JUMP, TRACK_CLICK_BEHAVIOR.STEP]);

var RAFLoop =
/*#__PURE__*/
function () {
  function RAFLoop() {
    var _this = this;

    _classCallCheck(this, RAFLoop);

    /**
     * @description List of targets to update
     */
    this.targets = [];
    /**
     * @description ID of requested animation frame. Valuable only if loop is active and has items to iterate.
     */

    this.animationFrameID = 0;
    /**
     * @description Loop's state.
     */

    this._isActive = false;
    /**
     * @description Start the loop if it wasn't yet.
     */

    this.start = function () {
      if (!_this._isActive && _this.targets.length) {
        _this._isActive = true;
        _this.animationFrameID && cancelAnimationFrame(_this.animationFrameID);
        _this.animationFrameID = requestAnimationFrame(_this.rafCallback);
      }

      return _this;
    };
    /**
     * @description Stop the loop if is was active.
     */


    this.stop = function () {
      if (_this._isActive) {
        _this._isActive = false;
        _this.animationFrameID && cancelAnimationFrame(_this.animationFrameID);
        _this.animationFrameID = 0;
      }

      return _this;
    };
    /**
     * @description Add target to the iteration list if it's not there.
     */


    this.addTarget = function (target) {
      var silent = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

      if (_this.targets.indexOf(target) === -1) {
        _this.targets.push(target);

        _this.targets.length === 1 && !silent && _this.start();
      }

      return _this;
    };
    /**
     * @description Remove target from iteration list if it was there.
     */


    this.removeTarget = function (target) {
      var idx = _this.targets.indexOf(target);

      if (idx !== -1) {
        _this.targets.splice(idx, 1);

        _this.targets.length === 0 && _this.stop();
      }

      return _this;
    };
    /**
     * @description Callback that called each animation frame.
     */


    this.rafCallback = function () {
      if (!_this._isActive) {
        return 0;
      }

      for (var i = 0; i < _this.targets.length; i++) {
        !_this.targets[i]._unmounted && _this.targets[i].update();
      }

      return _this.animationFrameID = requestAnimationFrame(_this.rafCallback);
    };
  }
  /**
   * @description Loop's state.
   */


  _createClass(RAFLoop, [{
    key: "isActive",
    get: function get() {
      return this._isActive;
    }
  }]);

  return RAFLoop;
}();
var Loop = new RAFLoop();

var ScrollbarTrack =
/*#__PURE__*/
function (_React$Component) {
  _inherits(ScrollbarTrack, _React$Component);

  function ScrollbarTrack() {
    var _this;

    _classCallCheck(this, ScrollbarTrack);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(ScrollbarTrack).apply(this, arguments));
    _this.element = null;

    _this.elementRef = function (ref) {
      typeof _this.props.elementRef === "function" && _this.props.elementRef(ref);
      _this.element = ref;
    };

    _this.handleClick = function (ev) {
      if (!ev || !_this.element || ev.button !== 0) {
        return;
      }

      if (typeof _this.props.onClick === "function" && ev.target === _this.element) {
        if (typeof ev.offsetX !== "undefined") {
          _this.props.onClick(ev, {
            axis: _this.props.axis,
            offset: _this.props.axis === AXIS_DIRECTION.X ? ev.offsetX : ev.offsetY
          });
        } else {
          // support for old browsers
          var rect = _this.element.getBoundingClientRect();

          _this.props.onClick(ev, {
            axis: _this.props.axis,
            offset: _this.props.axis === AXIS_DIRECTION.X ? ev.clientX - rect.left : ev.clientY - rect.top
          });
        }
      }

      return true;
    };

    return _this;
  }

  _createClass(ScrollbarTrack, [{
    key: "componentDidMount",
    value: function componentDidMount() {
      if (!this.element) {
        this.setState(function () {
          throw new Error("Element was not created. Possibly you haven't provided HTMLDivElement to renderer's `elementRef` function.");
        });
        return;
      }

      this.element.addEventListener("click", this.handleClick);
    }
  }, {
    key: "componentWillUnmount",
    value: function componentWillUnmount() {
      if (this.element) {
        this.element.removeEventListener("click", this.handleClick);
        this.element = null;
        this.elementRef(null);
      }
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props = this.props,
          elementRef = _this$props.elementRef,
          axis = _this$props.axis,
          onClick = _this$props.onClick,
          props = _objectWithoutProperties(_this$props, ["elementRef", "axis", "onClick"]);

      props.className = cnb("ScrollbarsCustom-Track", axis === AXIS_DIRECTION.X ? "ScrollbarsCustom-TrackX" : "ScrollbarsCustom-TrackY", props.className);

      if (props.renderer) {
        props.axis = axis;
      }

      return renderDivWithRenderer(props, this.elementRef);
    }
  }]);

  return ScrollbarTrack;
}(Component);
ScrollbarTrack.propTypes = {
  axis: AXIS_DIRECTION_PROP_TYPE,
  onClick: func,
  elementRef: func,
  renderer: func
};

var ScrollbarThumb =
/*#__PURE__*/
function (_React$Component) {
  _inherits(ScrollbarThumb, _React$Component);

  function ScrollbarThumb() {
    var _this;

    _classCallCheck(this, ScrollbarThumb);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(ScrollbarThumb).apply(this, arguments));
    _this.initialOffsetX = 0;
    _this.initialOffsetY = 0;
    _this.lastDragData = {
      x: 0,
      y: 0,
      deltaX: 0,
      deltaY: 0,
      lastX: 0,
      lastY: 0
    };
    _this.element = null;

    _this.handleOnDragStart = function (ev, data) {
      if (!_this.element) {
        _this.handleOnDragStop(ev, data);

        return;
      }

      if (global.document) {
        _this.prevUserSelect = global.document.body.style.userSelect;
        global.document.body.style.userSelect = "none";
        _this.prevOnSelectStart = global.document.onselectstart;
        global.document.onselectstart = ScrollbarThumb.selectStartReplacer;
      }

      _this.props.onDragStart && _this.props.onDragStart(_this.lastDragData = {
        x: data.x - _this.initialOffsetX,
        y: data.y - _this.initialOffsetY,
        lastX: data.lastX - _this.initialOffsetX,
        lastY: data.lastY - _this.initialOffsetY,
        deltaX: data.deltaX,
        deltaY: data.deltaY
      });

      _this.element.classList.add("dragging");
    };

    _this.handleOnDrag = function (ev, data) {
      if (!_this.element) {
        _this.handleOnDragStop(ev, data);

        return;
      }

      _this.props.onDrag && _this.props.onDrag(_this.lastDragData = {
        x: data.x - _this.initialOffsetX,
        y: data.y - _this.initialOffsetY,
        lastX: data.lastX - _this.initialOffsetX,
        lastY: data.lastY - _this.initialOffsetY,
        deltaX: data.deltaX,
        deltaY: data.deltaY
      });
    };

    _this.handleOnDragStop = function (ev, data) {
      var resultData = data ? {
        x: data.x - _this.initialOffsetX,
        y: data.y - _this.initialOffsetY,
        lastX: data.lastX - _this.initialOffsetX,
        lastY: data.lastY - _this.initialOffsetY,
        deltaX: data.deltaX,
        deltaY: data.deltaY
      } : _this.lastDragData;
      _this.props.onDragEnd && _this.props.onDragEnd(resultData);
      _this.element && _this.element.classList.remove("dragging");

      if (global.document) {
        global.document.body.style.userSelect = _this.prevUserSelect;
        _this.prevUserSelect = null;
        global.document.onselectstart = _this.prevOnSelectStart;
        _this.prevOnSelectStart = null;
      }

      _this.initialOffsetX = 0;
      _this.initialOffsetY = 0;
      _this.lastDragData = {
        x: 0,
        y: 0,
        deltaX: 0,
        deltaY: 0,
        lastX: 0,
        lastY: 0
      };
    };

    _this.handleOnMouseDown = function (ev) {
      if (!_this.element) {
        return;
      }

      ev.stopPropagation();

      if (typeof ev.offsetX !== "undefined") {
        _this.initialOffsetX = ev.offsetX;
        _this.initialOffsetY = ev.offsetY;
      } else {
        var rect = _this.element.getBoundingClientRect();

        _this.initialOffsetX = ev.clientX - rect.left;
        _this.initialOffsetY = ev.clientY - rect.top;
      }
    };

    _this.elementRef = function (ref) {
      typeof _this.props.elementRef === "function" && _this.props.elementRef(ref);
      _this.element = ref;
    };

    return _this;
  }

  _createClass(ScrollbarThumb, [{
    key: "componentDidMount",
    value: function componentDidMount() {
      if (!this.element) {
        this.setState(function () {
          throw new Error("Element was not created. Possibly you haven't provided HTMLDivElement to renderer's `elementRef` function.");
        });
        return;
      }
    }
  }, {
    key: "componentWillUnmount",
    value: function componentWillUnmount() {
      this.handleOnDragStop();
      this.elementRef(null);
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props = this.props,
          elementRef = _this$props.elementRef,
          axis = _this$props.axis,
          onDrag = _this$props.onDrag,
          onDragEnd = _this$props.onDragEnd,
          onDragStart = _this$props.onDragStart,
          props = _objectWithoutProperties(_this$props, ["elementRef", "axis", "onDrag", "onDragEnd", "onDragStart"]);

      props.className = cnb("ScrollbarsCustom-Thumb", axis === AXIS_DIRECTION.X ? "ScrollbarsCustom-ThumbX" : "ScrollbarsCustom-ThumbY", props.className);

      if (props.renderer) {
        props.axis = axis;
      }

      return createElement(DraggableCore, {
        allowAnyClick: false,
        enableUserSelectHack: false,
        onMouseDown: this.handleOnMouseDown,
        onDrag: this.handleOnDrag,
        onStart: this.handleOnDragStart,
        onStop: this.handleOnDragStop,
        children: renderDivWithRenderer(props, this.elementRef)
      });
    }
  }]);

  return ScrollbarThumb;
}(Component);
ScrollbarThumb.propTypes = {
  axis: AXIS_DIRECTION_PROP_TYPE,
  onDrag: func,
  onDragStart: func,
  onDragEnd: func,
  elementRef: func,
  renderer: func
};

ScrollbarThumb.selectStartReplacer = function () {
  return false;
};

var doc = global.document || null;
/**
 * @description Return element's height without padding
 *
 * ts-ignore here is okay here, because it brigs around 40% of performance
 */

var getInnerHeight = function getInnerHeight(el) {
  var styles = getComputedStyle(el);

  if (styles.boxSizing === "border-box") {
    return Math.max( // @ts-ignore
    (parseFloat(styles.height) || 0) - ( // @ts-ignore
    parseFloat(styles.paddingTop) || 0) - ( // @ts-ignore
    parseFloat(styles.paddingBottom) || 0), 0);
  } // @ts-ignore


  return parseFloat(styles.height) || 0;
};
/**
 * @description Return element's width without padding
 *
 * ts-ignore here is okay here, because it brigs around 40% of performance
 */

var getInnerWidth = function getInnerWidth(el) {
  var styles = getComputedStyle(el);

  if (styles.boxSizing === "border-box") {
    return Math.max( // @ts-ignore
    (parseFloat(styles.width) || 0) - ( // @ts-ignore
    parseFloat(styles.paddingLeft) || 0) - ( // @ts-ignore
    parseFloat(styles.paddingRight) || 0), 0);
  } // @ts-ignore


  return parseFloat(styles.width) || 0;
};
/**
 * @description Return unique UUID v4
 */

var uuid = function uuid() {
  var uuid = "";

  for (var i = 0; i < 32; i++) {
    if (i === 8 || i === 20) {
      uuid += "-" + (Math.random() * 16 | 0).toString(16);
    } else if (i === 12) {
      uuid += "-4";
    } else if (i === 16) {
      uuid += "-" + (Math.random() * 16 | 0 & 3 | 8).toString(16);
    } else {
      uuid += (Math.random() * 16 | 0).toString(16);
    }
  }

  return uuid;
};
/**
 * @description Calculate thumb size for given viewport and track parameters
 *
 * @param {number} contentSize - Scrollable content size
 * @param {number} viewportSize - Viewport size
 * @param {number} trackSize - Track size thumb can move
 * @param {number} minimalSize - Minimal thumb's size
 * @param {number} maximalSize - Maximal thumb's size
 */

function calcThumbSize(contentSize, viewportSize, trackSize, minimalSize, maximalSize) {
  if (viewportSize >= contentSize) {
    return 0;
  }

  var thumbSize = viewportSize / contentSize * trackSize;

  if (typeof maximalSize === "number") {
    thumbSize = Math.min(maximalSize, thumbSize);
  }

  if (typeof minimalSize === "number") {
    thumbSize = Math.max(minimalSize, thumbSize);
  }

  return thumbSize;
}
/**
 * @description Calculate thumb offset for given viewport, track and thumb parameters
 *
 * @param {number} contentSize - Scrollable content size
 * @param {number} viewportSize - Viewport size
 * @param {number} trackSize - Track size thumb can move
 * @param {number} thumbSize - Thumb size
 * @param {number} scroll - Scroll value to represent
 */

function calcThumbOffset(contentSize, viewportSize, trackSize, thumbSize, scroll) {
  if (!scroll || !thumbSize || viewportSize >= contentSize) {
    return 0;
  }

  return (trackSize - thumbSize) * scroll / (contentSize - viewportSize);
}
/**
 * @description Calculate scroll for given viewport, track and thumb parameters
 *
 * @param {number} contentSize - Scrollable content size
 * @param {number} viewportSize - Viewport size
 * @param {number} trackSize - Track size thumb can move
 * @param {number} thumbSize - Thumb size
 * @param {number} thumbOffset - Thumb's offset representing the scroll
 */

function calcScrollForThumbOffset(contentSize, viewportSize, trackSize, thumbSize, thumbOffset) {
  if (!thumbOffset || !thumbSize || viewportSize >= contentSize) {
    return 0;
  }

  return thumbOffset * (contentSize - viewportSize) / (trackSize - thumbSize);
}
var scrollbarWidth = null;
/**
 * @description Returns scrollbar width specific for current environment
 */

function getScrollbarWidth() {
  var force = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

  if (!force && scrollbarWidth !== null) {
    return scrollbarWidth;
  }

  if (!doc) {
    return scrollbarWidth = 0;
  }

  var el = doc.createElement("div");
  el.setAttribute("style", "display:block;position:absolute;width:100px;height:100px;top:-9999px;overflow:scroll;");
  doc.body.appendChild(el);
  scrollbarWidth = el.offsetWidth - el.clientWidth;
  doc.body.removeChild(el);
  return scrollbarWidth;
}
var isReverseRTLScrollNeeded = null;
/**
 * @description Detect need of horizontal scroll reverse while RTL
 */

var shouldReverseRTLScroll = function shouldReverseRTLScroll() {
  var force = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

  if (!force && isReverseRTLScrollNeeded !== null) {
    return isReverseRTLScrollNeeded;
  }

  if (!doc) {
    return isReverseRTLScrollNeeded = false;
  }

  var el = doc.createElement("div");
  var child = doc.createElement("div");
  el.setAttribute("style", "display:block;position:absolute;width:100px;height:100px;top:-9999px;overflow:scroll;direction:rtl;");
  child.setAttribute("style", "display:block;position:relative;width:1000px;height:1000px;direction:rtl;");
  el.appendChild(child);
  doc.body.appendChild(el);
  el.scrollLeft;
  el.scrollLeft = 45;
  isReverseRTLScrollNeeded = el.scrollLeft === 0;
  doc.body.removeChild(el);
  return isReverseRTLScrollNeeded;
};

var Emittr =
/*#__PURE__*/
function () {
  function Emittr() {
    var maxHandlers = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 10;

    _classCallCheck(this, Emittr);

    this.setMaxHandlers(maxHandlers);
    this._handlers = Object.create(null);
  }

  _createClass(Emittr, [{
    key: "setMaxHandlers",
    value: function setMaxHandlers(count) {
      if (typeof count !== "number" || count < 0 || !count) {
        throw new TypeError("Expected maxHandlers to be a positive number, got '".concat(count, "' of type ").concat(_typeof(count)));
      }

      this._maxHandlers = count;
      return this;
    }
  }, {
    key: "getMaxHandlers",
    value: function getMaxHandlers() {
      return this._maxHandlers;
    }
  }, {
    key: "emit",
    value: function emit(name) {
      if (_typeof(this._handlers[name]) !== "object" || !Array.isArray(this._handlers[name])) {
        return false;
      }

      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      Emittr._callEventHandlers(this, this._handlers[name], args);

      return true;
    }
  }, {
    key: "on",
    value: function on(name, handler) {
      Emittr._addHandler(this, name, handler);

      return this;
    }
  }, {
    key: "prependOn",
    value: function prependOn(name, handler) {
      Emittr._addHandler(this, name, handler, true);

      return this;
    }
  }, {
    key: "once",
    value: function once(name, handler) {
      if (typeof handler !== "function") {
        throw new TypeError("Expected event handler to be a function, got " + _typeof(handler));
      }

      Emittr._addHandler(this, name, this._wrapOnceHandler(name, handler));

      return this;
    }
  }, {
    key: "prependOnce",
    value: function prependOnce(name, handler) {
      if (typeof handler !== "function") {
        throw new TypeError("Expected event handler to be a function, got " + _typeof(handler));
      }

      Emittr._addHandler(this, name, this._wrapOnceHandler(name, handler), true);

      return this;
    }
  }, {
    key: "off",
    value: function off(name, handler) {
      Emittr._removeHandler(this, name, handler);

      return this;
    }
  }, {
    key: "removeAllHandlers",
    value: function removeAllHandlers() {
      var handlers = this._handlers;
      this._handlers = Object.create(null);
      var removeHandlers = handlers["removeHandler"];
      delete handlers["removeHandler"];
      var idx, eventName;

      for (eventName in handlers) {
        for (idx = handlers[eventName].length - 1; idx >= 0; idx--) {
          Emittr._callEventHandlers(this, removeHandlers, [eventName, handlers[eventName][idx].handler || handlers[eventName][idx]]);
        }
      }

      return this;
    }
  }, {
    key: "_wrapOnceHandler",
    value: function _wrapOnceHandler(name, handler) {
      var onceState = {
        fired: false,
        handler: handler,
        wrappedHandler: undefined,
        emitter: this,
        event: name
      };

      var wrappedHandler = Emittr._onceWrapper.bind(onceState);

      onceState.wrappedHandler = wrappedHandler;
      wrappedHandler.handler = handler;
      wrappedHandler.event = name;
      return wrappedHandler;
    }
  }], [{
    key: "_callEventHandlers",
    value: function _callEventHandlers(emitter, handlers, args) {
      if (!handlers.length) {
        return;
      }

      if (handlers.length === 1) {
        Reflect.apply(handlers[0], emitter, args);
        return;
      }

      handlers = _toConsumableArray(handlers);
      var idx;

      for (idx = 0; idx < handlers.length; idx++) {
        Reflect.apply(handlers[idx], emitter, args);
      }
    }
  }]);

  return Emittr;
}();

Emittr._addHandler = function (emitter, name, handler) {
  var prepend = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;

  if (typeof handler !== "function") {
    throw new TypeError("Expected event handler to be a function, got " + _typeof(handler));
  }

  emitter._handlers[name] = emitter._handlers[name] || [];
  emitter.emit("addHandler", name, handler);
  prepend ? emitter._handlers[name].unshift(handler) : emitter._handlers[name].push(handler);
  return emitter;
};

Emittr._onceWrapper = function _onceWrapper() {
  if (!this.fired) {
    this.fired = true;
    this.emitter.off(this.event, this.wrappedHandler);

    for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      args[_key2] = arguments[_key2];
    }

    Reflect.apply(this.handler, this.emitter, args);
  }
};

Emittr._removeHandler = function (emitter, name, handler) {
  if (typeof handler !== "function") {
    throw new TypeError("Expected event handler to be a function, got " + _typeof(handler));
  }

  if (typeof emitter._handlers[name] === "undefined" || !emitter._handlers[name].length) {
    return emitter;
  }

  var idx = -1;

  if (emitter._handlers[name].length === 1) {
    if (emitter._handlers[name][0] === handler || emitter._handlers[name][0].handler === handler) {
      idx = 0;
      handler = emitter._handlers[name][0].handler || emitter._handlers[name][0];
    }
  } else {
    for (idx = emitter._handlers[name].length - 1; idx >= 0; idx--) {
      if (emitter._handlers[name][idx] === handler || emitter._handlers[name][idx].handler === handler) {
        handler = emitter._handlers[name][idx].handler || emitter._handlers[name][idx];
        break;
      }
    }
  }

  if (idx === -1) {
    return emitter;
  }

  idx === 0 ? emitter._handlers[name].shift() : emitter._handlers[name].splice(idx, 1);
  emitter.emit("removeHandler", name, handler);
  return emitter;
};

var style = {
  holder: {
    position: "relative",
    width: "100%",
    height: "100%"
  },
  wrapper: {
    position: "absolute",
    top: 0,
    left: 0,
    bottom: 0,
    right: 0
  },
  content: {
    boxSizing: "border-box"
  },
  track: {
    common: {
      position: "absolute",
      overflow: "hidden",
      borderRadius: 4,
      background: "rgba(0,0,0,.1)",
      userSelect: "none"
    },
    x: {
      height: 10,
      width: "calc(100% - 20px)",
      bottom: 0,
      left: 10
    },
    y: {
      width: 10,
      height: "calc(100% - 20px)",
      top: 10
    }
  },
  thumb: {
    common: {
      cursor: "pointer",
      borderRadius: 4,
      background: "rgba(0,0,0,.4)"
    },
    x: {
      height: "100%",
      width: 0
    },
    y: {
      width: "100%",
      height: 0
    }
  }
};

var reverseRTL = shouldReverseRTLScroll();
var ScrollbarContext = createContext({
  parentScrollbar: null
});

var Scrollbar =
/*#__PURE__*/
function (_React$Component) {
  _inherits(Scrollbar, _React$Component);

  function Scrollbar(props) {
    var _this;

    _classCallCheck(this, Scrollbar);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Scrollbar).call(this, props));
    /**
     * @description Get current scroll-related values.<br/>
     * If <i>force</i> if truthy - will recalculate them instead of returning cached values.
     *
     * @return ScrollState
     */

    _this.getScrollState = function () {
      var force = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      if (_this.scrollValues && !force) {
        return _objectSpread({}, _this.scrollValues);
      }

      var scrollState = {
        clientHeight: 0,
        clientWidth: 0,
        contentScrollHeight: 0,
        contentScrollWidth: 0,
        scrollHeight: 0,
        scrollWidth: 0,
        scrollTop: 0,
        scrollLeft: 0,
        scrollYBlocked: false,
        scrollXBlocked: false,
        scrollYPossible: false,
        scrollXPossible: false,
        trackYVisible: false,
        trackXVisible: false,
        isRTL: undefined
      };
      scrollState.isRTL = _this.state.isRTL;
      scrollState.scrollYBlocked = _this.props.noScroll || _this.props.noScrollY;
      scrollState.scrollXBlocked = _this.props.noScroll || _this.props.noScrollX;

      if (_this.scrollerElement) {
        scrollState.clientHeight = _this.scrollerElement.clientHeight;
        scrollState.clientWidth = _this.scrollerElement.clientWidth;
        scrollState.scrollHeight = _this.scrollerElement.scrollHeight;
        scrollState.scrollWidth = _this.scrollerElement.scrollWidth;
        scrollState.scrollTop = _this.scrollerElement.scrollTop;
        scrollState.scrollLeft = _this.scrollerElement.scrollLeft;
        scrollState.scrollYPossible = !scrollState.scrollYBlocked && scrollState.scrollHeight > scrollState.clientHeight;
        scrollState.scrollXPossible = !scrollState.scrollXBlocked && scrollState.scrollWidth > scrollState.clientWidth;
        scrollState.trackYVisible = scrollState.scrollYPossible || _this.props.permanentTracks || _this.props.permanentTrackY;
        scrollState.trackXVisible = scrollState.scrollXPossible || _this.props.permanentTracks || _this.props.permanentTrackX;
      }

      if (_this.contentElement) {
        scrollState.contentScrollHeight = _this.contentElement.scrollHeight;
        scrollState.contentScrollWidth = _this.contentElement.scrollWidth;
      }

      return scrollState;
    };
    /**
     * @description Scroll to top border
     */


    _this.scrollToTop = function () {
      if (_this.scrollerElement) {
        _this.scrollerElement.scrollTop = 0;
      }

      return _assertThisInitialized(_this);
    };
    /**
     * @description Scroll to left border
     */


    _this.scrollToLeft = function () {
      if (_this.scrollerElement) {
        _this.scrollerElement.scrollLeft = 0;
      }

      return _assertThisInitialized(_this);
    };
    /**
     * @description Scroll to bottom border
     */


    _this.scrollToBottom = function () {
      if (_this.scrollerElement) {
        _this.scrollerElement.scrollTop = _this.scrollerElement.scrollHeight - _this.scrollerElement.clientHeight;
      }

      return _assertThisInitialized(_this);
    };
    /**
     * @description Scroll to right border
     */


    _this.scrollToRight = function () {
      if (_this.scrollerElement) {
        _this.scrollerElement.scrollLeft = _this.scrollerElement.scrollWidth - _this.scrollerElement.clientWidth;
      }

      return _assertThisInitialized(_this);
    };
    /**
     * @description Set the scrolls at given coordinates.<br/>
     * If coordinate is undefined - current scroll value will persist.
     */


    _this.scrollTo = function (x, y) {
      if (_this.scrollerElement) {
        typeof x === "number" && (_this.scrollerElement.scrollLeft = x);
        typeof y === "number" && (_this.scrollerElement.scrollTop = y);
      }

      return _assertThisInitialized(_this);
    };
    /**
     * @description Center the viewport at given coordinates.<br/>
     * If coordinate is undefined - current scroll value will persist.
     */


    _this.centerAt = function (x, y) {
      if (_this.scrollerElement) {
        typeof x === "number" && (_this.scrollerElement.scrollLeft = x - _this.scrollerElement.clientWidth / 2);
        typeof y === "number" && (_this.scrollerElement.scrollTop = y - _this.scrollerElement.clientHeight / 2);
      }

      return _assertThisInitialized(_this);
    };

    _this.update = function () {
      var force = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      if (!_this.scrollerElement) {
        return;
      } // autodetect direction if not defined


      if (typeof _this.state.isRTL === "undefined") {
        _this.setState({
          isRTL: getComputedStyle(_this.scrollerElement).direction === "rtl"
        });

        return _this.getScrollState();
      }

      var scrollState = _this.getScrollState(true);

      var prevScrollState = _objectSpread({}, _this.scrollValues);

      var bitmask = 0;

      if (!force) {
        prevScrollState.clientHeight !== scrollState.clientHeight && (bitmask |= 1 << 0);
        prevScrollState.clientWidth !== scrollState.clientWidth && (bitmask |= 1 << 1);
        prevScrollState.scrollHeight !== scrollState.scrollHeight && (bitmask |= 1 << 2);
        prevScrollState.scrollWidth !== scrollState.scrollWidth && (bitmask |= 1 << 3);
        prevScrollState.scrollTop !== scrollState.scrollTop && (bitmask |= 1 << 4);
        prevScrollState.scrollLeft !== scrollState.scrollLeft && (bitmask |= 1 << 5);
        prevScrollState.scrollYBlocked !== scrollState.scrollYBlocked && (bitmask |= 1 << 6);
        prevScrollState.scrollXBlocked !== scrollState.scrollXBlocked && (bitmask |= 1 << 7);
        prevScrollState.scrollYPossible !== scrollState.scrollYPossible && (bitmask |= 1 << 8);
        prevScrollState.scrollXPossible !== scrollState.scrollXPossible && (bitmask |= 1 << 9);
        prevScrollState.trackYVisible !== scrollState.trackYVisible && (bitmask |= 1 << 10);
        prevScrollState.trackXVisible !== scrollState.trackXVisible && (bitmask |= 1 << 11);
        prevScrollState.isRTL !== scrollState.isRTL && (bitmask |= 1 << 12);
        prevScrollState.contentScrollHeight !== scrollState.contentScrollHeight && (bitmask |= 1 << 13);
        prevScrollState.contentScrollWidth !== scrollState.contentScrollWidth && (bitmask |= 1 << 14); // if not forced and nothing has changed - skip this update

        if (bitmask === 0) {
          return prevScrollState;
        }
      } else {
        bitmask = 32767;
      }

      if (!_this.props.native && _this.holderElement) {
        if (bitmask & 1 << 13 && (_this.props.translateContentSizesToHolder || _this.props.translateContentSizeYToHolder)) {
          _this.holderElement.style.height = scrollState.contentScrollHeight + "px";
        }

        if (bitmask & 1 << 14 && (_this.props.translateContentSizesToHolder || _this.props.translateContentSizeXToHolder)) {
          _this.holderElement.style.width = scrollState.contentScrollWidth + "px";
        }

        if (_this.props.translateContentSizesToHolder || _this.props.translateContentSizeYToHolder || _this.props.translateContentSizeXToHolder) {
          if (!scrollState.clientHeight && scrollState.contentScrollHeight || !scrollState.clientWidth && scrollState.contentScrollWidth) {
            return;
          }
        }
      } // if scrollbars visibility has changed


      if (bitmask & 1 << 10 || bitmask & 1 << 11) {
        prevScrollState.scrollYBlocked = scrollState.scrollYBlocked;
        prevScrollState.scrollXBlocked = scrollState.scrollXBlocked;
        prevScrollState.scrollYPossible = scrollState.scrollYPossible;
        prevScrollState.scrollXPossible = scrollState.scrollXPossible;

        if (_this.trackYElement && bitmask & 1 << 10) {
          _this.trackYElement.style.display = scrollState.trackYVisible ? null : "none";
        }

        if (_this.trackXElement && bitmask & 1 << 11) {
          _this.trackXElement.style.display = scrollState.trackXVisible ? null : "none";
        }

        _this.scrollValues = prevScrollState;

        _this.setState({
          trackYVisible: _this.scrollValues.trackYVisible = scrollState.trackYVisible,
          trackXVisible: _this.scrollValues.trackXVisible = scrollState.trackXVisible
        });

        return;
      }

      (_this.props.native ? _this.updaterNative : _this.updaterCustom)(bitmask, scrollState);
      _this.scrollValues = scrollState;

      _this.eventEmitter.emit("update", _objectSpread({}, scrollState), prevScrollState);

      (bitmask & 1 << 4 || bitmask & 1 << 5) && _this.eventEmitter.emit("scroll", _objectSpread({}, scrollState), prevScrollState);
      return _this.scrollValues;
    };

    _this.updaterNative = function () {
      // just for future
      return true;
    };

    _this.updaterCustom = function (bitmask, scrollValues) {
      if (_this.trackYElement) {
        if (_this.thumbYElement && (bitmask & 1 << 0 || bitmask & 1 << 2 || bitmask & 1 << 4 || bitmask & 1 << 6 || bitmask & 1 << 8)) {
          if (scrollValues.scrollYPossible) {
            var trackInnerSize = getInnerHeight(_this.trackYElement);
            var thumbSize = calcThumbSize(scrollValues.scrollHeight, scrollValues.clientHeight, trackInnerSize, _this.props.minimalThumbYSize || _this.props.minimalThumbSize, _this.props.maximalThumbYSize || _this.props.maximalThumbSize);
            var thumbOffset = calcThumbOffset(scrollValues.scrollHeight, scrollValues.clientHeight, trackInnerSize, thumbSize, scrollValues.scrollTop);
            _this.thumbYElement.style.transform = "translateY(".concat(thumbOffset, "px)");
            _this.thumbYElement.style.height = thumbSize + "px";
            _this.thumbYElement.style.display = "";
          } else {
            _this.thumbYElement.style.transform = "";
            _this.thumbYElement.style.height = "0px";
            _this.thumbYElement.style.display = "none";
          }
        }
      }

      if (_this.trackXElement) {
        if (_this.thumbXElement && (bitmask & 1 << 1 || bitmask & 1 << 3 || bitmask & 1 << 5 || bitmask & 1 << 7 || bitmask & 1 << 9 || bitmask & 1 << 12)) {
          if (scrollValues.scrollXPossible) {
            var _trackInnerSize = getInnerWidth(_this.trackXElement);

            var _thumbSize = calcThumbSize(scrollValues.scrollWidth, scrollValues.clientWidth, _trackInnerSize, _this.props.minimalThumbXSize || _this.props.minimalThumbSize, _this.props.maximalThumbXSize || _this.props.maximalThumbSize);

            var _thumbOffset = calcThumbOffset(scrollValues.scrollWidth, scrollValues.clientWidth, _trackInnerSize, _thumbSize, scrollValues.scrollLeft);

            if (_this.state.isRTL && reverseRTL) {
              _thumbOffset += _trackInnerSize - _thumbSize;
            }

            _this.thumbXElement.style.transform = "translateX(".concat(_thumbOffset, "px)");
            _this.thumbXElement.style.width = _thumbSize + "px";
            _this.thumbXElement.style.display = "";
          } else {
            _this.thumbXElement.style.transform = "";
            _this.thumbXElement.style.width = "0px";
            _this.thumbXElement.style.display = "none";
          }
        }
      }

      return true;
    };

    _this.elementRefHolder = function (ref) {
      _this.holderElement = ref;
      typeof _this.props.elementRef === "function" && _this.props.elementRef(ref);
    };

    _this.elementRefWrapper = function (ref) {
      _this.wrapperElement = ref;
      typeof _this.props.wrapperProps.elementRef === "function" && _this.props.wrapperProps.elementRef(ref);
    };

    _this.elementRefScroller = function (ref) {
      _this.scrollerElement = ref;
      typeof _this.props.scrollerProps.elementRef === "function" && _this.props.scrollerProps.elementRef(ref);
    };

    _this.elementRefContent = function (ref) {
      _this.contentElement = ref;
      typeof _this.props.contentProps.elementRef === "function" && _this.props.contentProps.elementRef(ref);
    };

    _this.elementRefTrackX = function (ref) {
      _this.trackXElement = ref;
      typeof _this.props.trackXProps.elementRef === "function" && _this.props.trackXProps.elementRef(ref);
    };

    _this.elementRefTrackY = function (ref) {
      _this.trackYElement = ref;
      typeof _this.props.trackYProps.elementRef === "function" && _this.props.trackYProps.elementRef(ref);
    };

    _this.elementRefThumbX = function (ref) {
      _this.thumbXElement = ref;
      typeof _this.props.thumbXProps.elementRef === "function" && _this.props.thumbXProps.elementRef(ref);
    };

    _this.elementRefThumbY = function (ref) {
      _this.thumbYElement = ref;
      typeof _this.props.thumbYProps.elementRef === "function" && _this.props.thumbYProps.elementRef(ref);
    };

    _this.handleTrackXClick = function (ev, values) {
      _this.props.trackXProps.onClick && _this.props.trackXProps.onClick(ev, values);

      if (!_this.scrollerElement || !_this.trackXElement || !_this.thumbXElement || !_this.scrollValues || !_this.scrollValues.scrollXPossible) {
        return;
      }

      _this._scrollDetection();

      var thumbSize = _this.thumbXElement.clientWidth;
      var trackInnerSize = getInnerWidth(_this.trackXElement);
      var thumbOffset = (_this.scrollValues.isRTL && reverseRTL ? values.offset + thumbSize / 2 - trackInnerSize : values.offset - thumbSize / 2) - ( //@ts-ignore
      parseFloat(getComputedStyle(_this.trackXElement).paddingLeft) || 0);
      var target = calcScrollForThumbOffset(_this.scrollValues.scrollWidth, _this.scrollValues.clientWidth, trackInnerSize, thumbSize, thumbOffset);

      if (_this.props.trackClickBehavior === TRACK_CLICK_BEHAVIOR.STEP) {
        target = (_this.scrollValues.isRTL ? _this.scrollValues.scrollLeft > target : _this.scrollValues.scrollLeft < target) ? _this.scrollValues.scrollLeft + _this.scrollValues.clientWidth : _this.scrollValues.scrollLeft - _this.scrollValues.clientWidth;
      }

      _this.scrollerElement.scrollLeft = target;
    };

    _this.handleTrackYClick = function (ev, values) {
      _this.props.trackYProps.onClick && _this.props.trackYProps.onClick(ev, values);

      if (!_this.scrollerElement || !_this.trackYElement || !_this.thumbYElement || !_this.scrollValues || !_this.scrollValues.scrollYPossible) {
        return;
      }

      _this._scrollDetection();

      var thumbSize = _this.thumbYElement.clientHeight;
      var target = calcScrollForThumbOffset(_this.scrollValues.scrollHeight, _this.scrollValues.clientHeight, getInnerHeight(_this.trackYElement), thumbSize, values.offset - thumbSize / 2) - ( //@ts-ignore
      parseFloat(getComputedStyle(_this.trackYElement).paddingTop) || 0);

      if (_this.props.trackClickBehavior === TRACK_CLICK_BEHAVIOR.JUMP) {
        _this.scrollerElement.scrollTop = target;
      } else {
        _this.scrollerElement.scrollTop = _this.scrollValues.scrollTop < target ? _this.scrollValues.scrollTop + _this.scrollValues.clientHeight : _this.scrollValues.scrollTop - _this.scrollValues.clientHeight;
      }
    };

    _this.handleTrackYMouseWheel = function (ev) {
      _this.props.trackYProps && _this.props.trackYProps.onWheel && _this.props.trackYProps.onWheel(ev);

      if (_this.props.disableTracksMousewheelScrolling || _this.props.disableTrackYMousewheelScrolling) {
        return;
      }

      _this._scrollDetection();

      if (!_this.scrollerElement || _this.scrollValues.scrollYBlocked) {
        return;
      }

      _this.scrollTop += ev.deltaY;
    };

    _this.handleTrackXMouseWheel = function (ev) {
      _this.props.trackXProps && _this.props.trackXProps.onWheel && _this.props.trackXProps.onWheel(ev);

      if (_this.props.disableTracksMousewheelScrolling || _this.props.disableTrackXMousewheelScrolling) {
        return;
      }

      _this._scrollDetection();

      if (!_this.scrollerElement || _this.scrollValues.scrollXBlocked) {
        return;
      }

      _this.scrollLeft += ev.deltaX;
    };

    _this.handleThumbXDrag = function (data) {
      if (!_this.trackXElement || !_this.thumbXElement || !_this.scrollerElement || !_this.scrollValues || !_this.scrollValues.scrollXPossible) {
        return;
      }

      _this._scrollDetection();

      var trackRect = _this.trackXElement.getBoundingClientRect();

      var styles = getComputedStyle(_this.trackXElement); //@ts-ignore

      var paddingLeft = parseFloat(styles.paddingLeft) || 0; //@ts-ignore

      var paddingRight = parseFloat(styles.paddingRight) || 0;
      var trackInnerSize = trackRect.width - paddingLeft - paddingRight;
      var thumbSize = _this.thumbXElement.clientWidth;
      var offset = _this.scrollValues.isRTL && reverseRTL ? data.x + thumbSize - trackInnerSize + paddingLeft : data.lastX - paddingLeft;
      _this.scrollerElement.scrollLeft = calcScrollForThumbOffset(_this.scrollValues.scrollWidth, _this.scrollValues.clientWidth, trackInnerSize, thumbSize, offset);
    };

    _this.handleThumbYDrag = function (data) {
      if (!_this.scrollerElement || !_this.trackYElement || !_this.thumbYElement || !_this.scrollValues || !_this.scrollValues.scrollYPossible) {
        return;
      }

      _this._scrollDetection();

      var trackRect = _this.trackYElement.getBoundingClientRect();

      var styles = getComputedStyle(_this.trackYElement); //@ts-ignore

      var paddingTop = parseFloat(styles.paddingTop) || 0; //@ts-ignore

      var paddingBottom = parseFloat(styles.paddingBottom) || 0;
      var trackInnerSize = trackRect.height - paddingTop - paddingBottom;
      var thumbSize = _this.thumbYElement.clientHeight;
      var offset = data.y - paddingTop;
      _this.scrollerElement.scrollTop = calcScrollForThumbOffset(_this.scrollValues.scrollHeight, _this.scrollValues.clientHeight, trackInnerSize, thumbSize, offset);
    };

    _this.handleScrollerScroll = function () {
      _this._scrollDetection();
    };

    _this._scrollDetection = function () {
      !_this._scrollDetectionTO && _this.eventEmitter.emit("scrollStart", _this.getScrollState());
      _this._scrollDetectionTO && global.window && global.window.clearTimeout(_this._scrollDetectionTO);
      _this._scrollDetectionTO = global.window ? global.window.setTimeout(_this._scrollDetectionCallback, _this.props.scrollDetectionThreshold || 0) : null;
    };

    _this._scrollDetectionCallback = function () {
      _this._scrollDetectionTO = null;

      _this.eventEmitter.emit("scrollStop", _this.getScrollState());
    };

    _this.state = {
      trackXVisible: false,
      trackYVisible: false,
      isRTL: _this.props.rtl
    };
    _this.scrollValues = _this.getScrollState(true);
    _this.eventEmitter = new Emittr(15);
    _this.props.onUpdate && _this.eventEmitter.on("update", _this.props.onUpdate);
    _this.props.onScroll && _this.eventEmitter.on("scroll", _this.props.onScroll);
    _this.props.onScrollStart && _this.eventEmitter.on("scrollStart", _this.props.onScrollStart);
    _this.props.onScrollStop && _this.eventEmitter.on("scrollStop", _this.props.onScrollStop);
    _this.id = uuid();
    return _this;
  }

  _createClass(Scrollbar, [{
    key: "componentDidMount",
    value: function componentDidMount() {
      if (!this.scrollerElement) {
        this.setState(function () {
          throw new Error("scroller element was not created. Possibly you haven't provided HTMLDivElement to renderer's `elementRef` function.");
        });
        return;
      }

      if (!this.contentElement) {
        this.setState(function () {
          throw new Error("content element was not created. Possibly you haven't provided HTMLDivElement to renderer's `elementRef` function.");
        });
        return;
      }

      if (!this.props.native && !this.props.mobileNative) {
        //ToDo: move native state to the state so it can be synchronized
        if (!this.holderElement) {
          this.setState(function () {
            throw new Error("holder element was not created. Possibly you haven't provided HTMLDivElement to renderer's `elementRef` function.");
          });
          return;
        }

        if (!this.wrapperElement) {
          this.setState(function () {
            throw new Error("wrapper element was not created. Possibly you haven't provided HTMLDivElement to renderer's `elementRef` function.");
          });
          return;
        }
      }

      Loop.addTarget(this);

      if (typeof this.props.scrollTop !== "undefined") {
        this.scrollerElement.scrollTop = this.props.scrollTop;
      }

      if (typeof this.props.scrollLeft !== "undefined") {
        this.scrollerElement.scrollLeft = this.props.scrollLeft;
      }

      this.update(true);
    }
  }, {
    key: "componentWillUnmount",
    value: function componentWillUnmount() {
      Loop.removeTarget(this);
    }
  }, {
    key: "componentDidUpdate",
    value: function componentDidUpdate(prevProps, prevState) {
      if (!this.scrollerElement) {
        return;
      }

      if (this.props.rtl !== prevProps.rtl && this.props.rtl !== this.state.isRTL) {
        this.setState({
          isRTL: this.props.rtl
        });
      }

      if (this.state.isRTL !== prevState.isRTL) {
        this.update();
      }

      if (typeof this.props.scrollTop !== "undefined" && this.props.scrollTop !== this.scrollerElement.scrollTop) {
        this.scrollerElement.scrollTop = this.props.scrollTop;
      }

      if (typeof this.props.scrollLeft !== "undefined" && this.props.scrollLeft !== this.scrollerElement.scrollLeft) {
        this.scrollerElement.scrollLeft = this.props.scrollLeft;
      }

      if (prevProps.onUpdate !== this.props.onUpdate) {
        prevProps.onUpdate && this.eventEmitter.off("update", prevProps.onUpdate);
        this.props.onUpdate && this.eventEmitter.on("update", this.props.onUpdate);
      }

      if (prevProps.onScroll !== this.props.onScroll) {
        prevProps.onScroll && this.eventEmitter.off("scroll", prevProps.onScroll);
        this.props.onScroll && this.eventEmitter.on("scroll", this.props.onScroll);
      }

      if (prevProps.onScrollStart !== this.props.onScrollStart) {
        prevProps.onScrollStart && this.eventEmitter.off("scrollStart", prevProps.onScrollStart);
        this.props.onScrollStart && this.eventEmitter.on("scrollStart", this.props.onScrollStart);
      }

      if (prevProps.onScrollStop !== this.props.onScrollStop) {
        prevProps.onScrollStop && this.eventEmitter.off("scrollStop", prevProps.onScrollStop);
        this.props.onScrollStop && this.eventEmitter.on("scrollStop", this.props.onScrollStop);
      }
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props = this.props,
          createContext = _this$props.createContext,
          rtl = _this$props.rtl,
          native = _this$props.native,
          mobileNative = _this$props.mobileNative,
          momentum = _this$props.momentum,
          noDefaultStyles = _this$props.noDefaultStyles,
          disableTracksMousewheelScrolling = _this$props.disableTracksMousewheelScrolling,
          disableTrackXMousewheelScrolling = _this$props.disableTrackXMousewheelScrolling,
          disableTrackYMousewheelScrolling = _this$props.disableTrackYMousewheelScrolling,
          disableTracksWidthCompensation = _this$props.disableTracksWidthCompensation,
          disableTrackXWidthCompensation = _this$props.disableTrackXWidthCompensation,
          disableTrackYWidthCompensation = _this$props.disableTrackYWidthCompensation,
          noScrollX = _this$props.noScrollX,
          noScrollY = _this$props.noScrollY,
          noScroll = _this$props.noScroll,
          permanentTrackX = _this$props.permanentTrackX,
          permanentTrackY = _this$props.permanentTrackY,
          permanentTracks = _this$props.permanentTracks,
          removeTracksWhenNotUsed = _this$props.removeTracksWhenNotUsed,
          removeTrackYWhenNotUsed = _this$props.removeTrackYWhenNotUsed,
          removeTrackXWhenNotUsed = _this$props.removeTrackXWhenNotUsed,
          minimalThumbSize = _this$props.minimalThumbSize,
          maximalThumbSize = _this$props.maximalThumbSize,
          minimalThumbXSize = _this$props.minimalThumbXSize,
          maximalThumbXSize = _this$props.maximalThumbXSize,
          minimalThumbYSize = _this$props.minimalThumbYSize,
          maximalThumbYSize = _this$props.maximalThumbYSize,
          fallbackScrollbarWidth = _this$props.fallbackScrollbarWidth,
          scrollTop = _this$props.scrollTop,
          scrollLeft = _this$props.scrollLeft,
          trackClickBehavior = _this$props.trackClickBehavior,
          scrollDetectionThreshold = _this$props.scrollDetectionThreshold,
          propsWrapperProps = _this$props.wrapperProps,
          propsScrollerProps = _this$props.scrollerProps,
          propsContentProps = _this$props.contentProps,
          propsTrackXProps = _this$props.trackXProps,
          propsTrackYProps = _this$props.trackYProps,
          propsThumbXProps = _this$props.thumbXProps,
          propsThumbYProps = _this$props.thumbYProps,
          propsScrollbarWidth = _this$props.scrollbarWidth,
          elementRef = _this$props.elementRef,
          onUpdate = _this$props.onUpdate,
          onScroll = _this$props.onScroll,
          onScrollStart = _this$props.onScrollStart,
          onScrollStop = _this$props.onScrollStop,
          translateContentSizesToHolder = _this$props.translateContentSizesToHolder,
          translateContentSizeYToHolder = _this$props.translateContentSizeYToHolder,
          translateContentSizeXToHolder = _this$props.translateContentSizeXToHolder,
          children = _this$props.children,
          propsHolderProps = _objectWithoutProperties(_this$props, ["createContext", "rtl", "native", "mobileNative", "momentum", "noDefaultStyles", "disableTracksMousewheelScrolling", "disableTrackXMousewheelScrolling", "disableTrackYMousewheelScrolling", "disableTracksWidthCompensation", "disableTrackXWidthCompensation", "disableTrackYWidthCompensation", "noScrollX", "noScrollY", "noScroll", "permanentTrackX", "permanentTrackY", "permanentTracks", "removeTracksWhenNotUsed", "removeTrackYWhenNotUsed", "removeTrackXWhenNotUsed", "minimalThumbSize", "maximalThumbSize", "minimalThumbXSize", "maximalThumbXSize", "minimalThumbYSize", "maximalThumbYSize", "fallbackScrollbarWidth", "scrollTop", "scrollLeft", "trackClickBehavior", "scrollDetectionThreshold", "wrapperProps", "scrollerProps", "contentProps", "trackXProps", "trackYProps", "thumbXProps", "thumbYProps", "scrollbarWidth", "elementRef", "onUpdate", "onScroll", "onScrollStart", "onScrollStop", "translateContentSizesToHolder", "translateContentSizeYToHolder", "translateContentSizeXToHolder", "children"]);

      var scrollbarWidth = typeof propsScrollbarWidth !== "undefined" ? propsScrollbarWidth : getScrollbarWidth() || 0;

      if (native || !scrollbarWidth && mobileNative) {
        this.elementRefHolder(null);
        this.elementRefWrapper(null);
        this.elementRefTrackX(null);
        this.elementRefTrackY(null);
        this.elementRefThumbX(null);
        this.elementRefThumbY(null);

        var _contentProps = _objectSpread({}, this.props.contentProps, {
          key: "ScrollbarsCustom-Content",
          className: cnb("ScrollbarsCustom-Content", propsContentProps.className),
          children: this.props.children
        });

        var _scrollerProps = _objectSpread({}, propsHolderProps, {
          className: cnb("ScrollbarsCustom native", this.state.trackYVisible && "trackYVisible", this.state.trackXVisible && "trackXVisible", this.state.isRTL && "rtl", propsHolderProps.className),
          style: _objectSpread({}, propsHolderProps.style, typeof rtl !== "undefined" && {
            direction: rtl ? "rtl" : "ltr"
          }, momentum && {
            WebkitOverflowScrolling: "touch"
          }, {
            overflowX: noScroll || noScrollX ? "hidden" : permanentTracks || permanentTrackX ? "scroll" : "auto",
            overflowY: noScroll || noScrollY ? "hidden" : permanentTracks || permanentTrackY ? "scroll" : "auto"
          }),
          onScroll: this.handleScrollerScroll,
          children: renderDivWithRenderer(_contentProps, this.elementRefContent),
          renderer: propsScrollerProps.renderer,
          elementRef: propsScrollerProps.elementRef
        });

        return renderDivWithRenderer(_scrollerProps, this.elementRefScroller);
      }

      var styles = Scrollbar.calculateStyles(this.props, this.state, this.scrollValues, scrollbarWidth);
      var holderChildren = [];

      var contentProps = _objectSpread({}, propsContentProps, {
        key: "ScrollbarsCustom-Content",
        className: cnb("ScrollbarsCustom-Content", propsContentProps.className),
        style: styles.content,
        children: createContext ? createElement(ScrollbarContext.Provider, {
          value: {
            parentScrollbar: this
          },
          children: children
        }) : children
      });

      var scrollerProps = _objectSpread({}, propsScrollerProps, {
        key: "ScrollbarsCustom-Scroller",
        className: cnb("ScrollbarsCustom-Scroller", propsScrollerProps.className),
        style: styles.scroller,
        children: renderDivWithRenderer(contentProps, this.elementRefContent),
        onScroll: this.handleScrollerScroll
      });

      var wrapperProps = _objectSpread({}, propsWrapperProps, {
        key: "ScrollbarsCustom-Wrapper",
        className: cnb("ScrollbarsCustom-Wrapper", propsWrapperProps.className),
        style: styles.wrapper,
        children: renderDivWithRenderer(scrollerProps, this.elementRefScroller)
      });

      holderChildren.push(renderDivWithRenderer(wrapperProps, this.elementRefWrapper));

      if (this.state.trackYVisible || !removeTracksWhenNotUsed && !removeTrackYWhenNotUsed) {
        var thumbYProps = _objectSpread({}, propsThumbYProps, {
          key: "ScrollbarsCustom-ThumbY",
          style: styles.thumbY,
          elementRef: this.elementRefThumbY,
          onDrag: this.handleThumbYDrag,
          onDragEnd: this.handleThumbYDrag,
          axis: AXIS_DIRECTION.Y
        });

        var trackYProps = _objectSpread({}, propsTrackYProps, {
          key: "ScrollbarsCustom-TrackY",
          style: styles.trackY,
          elementRef: this.elementRefTrackY,
          onClick: this.handleTrackYClick
        }, (disableTracksMousewheelScrolling || disableTrackYMousewheelScrolling) && {
          onWheel: this.handleTrackYMouseWheel
        }, {
          axis: AXIS_DIRECTION.Y
        });

        trackYProps.children = createElement(ScrollbarThumb, Object.assign({}, thumbYProps));
        holderChildren.push(createElement(ScrollbarTrack, Object.assign({}, trackYProps)));
      } else {
        this.elementRefTrackY(null);
        this.elementRefThumbY(null);
      }

      if (this.state.trackXVisible || !removeTracksWhenNotUsed && !removeTrackXWhenNotUsed) {
        var thumbXProps = _objectSpread({}, propsThumbXProps, {
          key: "ScrollbarsCustom-ThumbX",
          style: styles.thumbX,
          elementRef: this.elementRefThumbX,
          onDrag: this.handleThumbXDrag,
          onDragEnd: this.handleThumbXDrag,
          axis: AXIS_DIRECTION.X
        });

        var trackXProps = _objectSpread({}, propsTrackXProps, {
          key: "ScrollbarsCustom-TrackX",
          style: styles.trackX,
          elementRef: this.elementRefTrackX,
          onClick: this.handleTrackXClick
        }, (disableTracksMousewheelScrolling || disableTrackXMousewheelScrolling) && {
          onWheel: this.handleTrackXMouseWheel
        }, {
          axis: AXIS_DIRECTION.X
        });

        trackXProps.children = createElement(ScrollbarThumb, Object.assign({}, thumbXProps));
        holderChildren.push(createElement(ScrollbarTrack, Object.assign({}, trackXProps)));
      } else {
        this.elementRefTrackX(null);
        this.elementRefThumbX(null);
      }

      var holderProps = _objectSpread({}, propsHolderProps, {
        className: cnb("ScrollbarsCustom", this.state.trackYVisible && "trackYVisible", this.state.trackXVisible && "trackXVisible", this.state.isRTL && "rtl", propsHolderProps.className),
        style: styles.holder,
        children: holderChildren
      });

      return renderDivWithRenderer(holderProps, this.elementRefHolder);
    }
  }, {
    key: "scrollTop",
    get: function get() {
      if (this.scrollerElement) {
        return this.scrollerElement.scrollTop;
      }

      return 0;
    },
    set: function set(top) {
      if (this.scrollerElement) {
        this.scrollerElement.scrollTop = top;
        this.update();
      }
    }
  }, {
    key: "scrollLeft",
    get: function get() {
      if (this.scrollerElement) {
        return this.scrollerElement.scrollLeft;
      }

      return 0;
    },
    set: function set(left) {
      if (this.scrollerElement) {
        this.scrollerElement.scrollLeft = left;
      }
    }
  }, {
    key: "scrollHeight",
    get: function get() {
      if (this.scrollerElement) {
        return this.scrollerElement.scrollHeight;
      }

      return 0;
    }
  }, {
    key: "scrollWidth",
    get: function get() {
      if (this.scrollerElement) {
        return this.scrollerElement.scrollWidth;
      }

      return 0;
    }
  }, {
    key: "clientHeight",
    get: function get() {
      if (this.scrollerElement) {
        return this.scrollerElement.clientHeight;
      }

      return 0;
    }
  }, {
    key: "clientWidth",
    get: function get() {
      if (this.scrollerElement) {
        return this.scrollerElement.clientWidth;
      }

      return 0;
    }
  }], [{
    key: "calculateStyles",
    value: function calculateStyles(props, state, scrollValues, scrollbarWidth) {
      var useDefaultStyles = !props.noDefaultStyles;
      return {
        holder: _objectSpread({}, useDefaultStyles && style.holder, {
          position: "relative"
        }, props.style),
        wrapper: _objectSpread({}, useDefaultStyles && _objectSpread({}, style.wrapper, !props.disableTracksWidthCompensation && !props.disableTrackYWidthCompensation && _defineProperty({}, state.isRTL ? "left" : "right", state.trackYVisible ? 10 : 0), !props.disableTracksWidthCompensation && !props.disableTrackXWidthCompensation && {
          bottom: state.trackXVisible ? 10 : 0
        }), props.wrapperProps.style, {
          position: "absolute",
          overflow: "hidden"
        }),
        content: _objectSpread({}, useDefaultStyles && style.content, props.translateContentSizesToHolder || props.translateContentSizeYToHolder || props.translateContentSizeXToHolder ? {
          display: "table-cell"
        } : {
          padding: 0.05 // needed to disable margin collapsing without flexboxes, other possible solutions here: https://stackoverflow.com/questions/19718634/how-to-disable-margin-collapsing

        }, useDefaultStyles && !(props.translateContentSizesToHolder || props.translateContentSizeYToHolder) && {
          minHeight: "100%"
        }, useDefaultStyles && !(props.translateContentSizesToHolder || props.translateContentSizeXToHolder) && {
          minWidth: "100%"
        }, props.contentProps.style),
        scroller: _objectSpread(_defineProperty({
          position: "absolute",
          top: 0,
          left: 0,
          bottom: 0,
          right: 0,
          paddingBottom: !scrollbarWidth && scrollValues.scrollXPossible ? props.fallbackScrollbarWidth : undefined
        }, state.isRTL ? "paddingLeft" : "paddingRight", !scrollbarWidth && scrollValues.scrollYPossible ? props.fallbackScrollbarWidth : undefined), props.scrollerProps.style, typeof props.rtl !== "undefined" && {
          direction: props.rtl ? "rtl" : "ltr"
        }, props.momentum && {
          WebkitOverflowScrolling: "touch"
        }, _defineProperty({
          overflowY: scrollValues.scrollYPossible ? "scroll" : "hidden",
          overflowX: scrollValues.scrollXPossible ? "scroll" : "hidden",
          marginBottom: scrollValues.scrollXPossible ? -(scrollbarWidth || props.fallbackScrollbarWidth) : undefined
        }, state.isRTL ? "marginLeft" : "marginRight", scrollValues.scrollYPossible ? -(scrollbarWidth || props.fallbackScrollbarWidth) : undefined)),
        trackX: _objectSpread({}, useDefaultStyles && style.track.common, useDefaultStyles && style.track.x, props.trackXProps.style, !state.trackXVisible && {
          display: "none"
        }),
        trackY: _objectSpread({}, useDefaultStyles && style.track.common, useDefaultStyles && style.track.y, useDefaultStyles && _defineProperty({}, state.isRTL ? "left" : "right", 0), props.trackYProps.style, !state.trackYVisible && {
          display: "none"
        }),
        thumbX: _objectSpread({}, useDefaultStyles && style.thumb.common, useDefaultStyles && style.thumb.x, props.thumbXProps.style),
        thumbY: _objectSpread({}, useDefaultStyles && style.thumb.common, useDefaultStyles && style.thumb.y, props.thumbYProps.style)
      };
    }
  }]);

  return Scrollbar;
}(Component);
Scrollbar.contextType = ScrollbarContext;
Scrollbar.propTypes = {
  createContext: bool,
  rtl: bool,
  native: bool,
  mobileNative: bool,
  momentum: bool,
  noDefaultStyles: bool,
  disableTracksMousewheelScrolling: bool,
  disableTrackXMousewheelScrolling: bool,
  disableTrackYMousewheelScrolling: bool,
  disableTracksWidthCompensation: bool,
  disableTrackXWidthCompensation: bool,
  disableTrackYWidthCompensation: bool,
  minimalThumbSize: number,
  maximalThumbSize: number,
  minimalThumbXSize: number,
  maximalThumbXSize: number,
  minimalThumbYSize: number,
  maximalThumbYSize: number,
  noScrollX: bool,
  noScrollY: bool,
  noScroll: bool,
  permanentTrackX: bool,
  permanentTrackY: bool,
  permanentTracks: bool,
  translateContentSizesToHolder: bool,
  translateContentSizeYToHolder: bool,
  translateContentSizeXToHolder: bool,
  removeTracksWhenNotUsed: bool,
  removeTrackYWhenNotUsed: bool,
  removeTrackXWhenNotUsed: bool,
  trackClickBehavior: TRACK_CLICK_BEHAVIOR_PROP_TYPE,
  scrollbarWidth: number,
  fallbackScrollbarWidth: number,
  scrollDetectionThreshold: number,
  scrollTop: number,
  scrollLeft: number,
  className: string,
  wrapperProps: object,
  contentProps: object,
  trackXProps: object,
  trackYProps: object,
  thumbXProps: object,
  thumbYProps: object,
  onUpdate: func,
  onScroll: func,
  onScrollStart: func,
  onScrollStop: func
};
Scrollbar.defaultProps = {
  momentum: true,
  minimalThumbSize: 30,
  fallbackScrollbarWidth: 20,
  trackClickBehavior: TRACK_CLICK_BEHAVIOR.JUMP,
  scrollDetectionThreshold: 100,
  wrapperProps: {},
  scrollerProps: {},
  contentProps: {},
  trackXProps: {},
  trackYProps: {},
  thumbXProps: {},
  thumbYProps: {}
};

export default Scrollbar;
export { Scrollbar, ScrollbarContext };
